@@ -1,5 +1,5 @@
 /* Stylesheet.java -- 
-   Copyright (C) 2004 Free Software Foundation, Inc.
+   Copyright (C) 2004,2006 Free Software Foundation, Inc.
 
 This file is part of GNU Classpath.
 
@@ -37,6 +37,8 @@ exception statement from your version. */
 
 package gnu.xml.transform;
 
+import gnu.java.lang.CPStringBuilder;
+
 import java.text.DecimalFormat;
 import java.text.DecimalFormatSymbols;
 import java.util.ArrayList;
@@ -51,7 +53,6 @@ import java.util.List;
 import java.util.Map;
 import java.util.Set;
 import java.util.StringTokenizer;
-import java.util.TreeSet;
 import javax.xml.XMLConstants;
 import javax.xml.namespace.NamespaceContext;
 import javax.xml.namespace.QName;
@@ -88,6 +89,8 @@ class Stylesheet
 {
 
   static final String XSL_NS = "http://www.w3.org/1999/XSL/Transform";
+  private static final NameTest STYLESHEET_PRESERVE_TEXT =
+    new NameTest(new QName(XSL_NS, "text"), false, false);
   
   static final int OUTPUT_XML = 0;
   static final int OUTPUT_HTML = 1;
@@ -107,18 +110,18 @@ class Stylesheet
    */
   String version;
 
-  Collection extensionElementPrefixes;
-  Collection excludeResultPrefixes;
+  Collection<String> extensionElementPrefixes;
+  Collection<String> excludeResultPrefixes;
 
   /**
    * Set of element names for which we should strip whitespace.
    */
-  Set stripSpace;
+  Set<StrippingInstruction> stripSpace;
 
   /**
    * Set of element names for which we should preserve whitespace.
    */
-  Set preserveSpace;
+  Set<StrippingInstruction> preserveSpace;
 
   /**
    * Output options.
@@ -131,34 +134,34 @@ class Stylesheet
   boolean outputStandalone;
   String outputPublicId;
   String outputSystemId;
-  Collection outputCdataSectionElements;
+  Collection<String> outputCdataSectionElements;
   boolean outputIndent;
   String outputMediaType;
 
   /**
    * Keys.
    */
-  Collection keys;
+  Collection<Key> keys;
 
   /**
    * Decimal formats.
    */
-  Map decimalFormats;
+  Map<String,DecimalFormat> decimalFormats;
   
   /**
    * Namespace aliases.
    */
-  Map namespaceAliases;
+  Map<String,String> namespaceAliases;
 
   /**
    * Attribute-sets.
    */
-  List attributeSets;
+  List<AttributeSet> attributeSets;
 
   /**
    * Variables.
    */
-  List variables;
+  List<ParameterNode> variables;
 
   /**
    * Variable and parameter bindings.
@@ -168,7 +171,7 @@ class Stylesheet
   /**
    * Templates.
    */
-  LinkedList templates;
+  LinkedList<Template> templates;
 
   TemplateNode builtInNodeTemplate;
   TemplateNode builtInTextTemplate;
@@ -202,23 +205,23 @@ class Stylesheet
     this.systemId = systemId;
     this.precedence = precedence;
     this.parent = parent;
-    extensionElementPrefixes = new HashSet();
-    excludeResultPrefixes = new HashSet();
-    stripSpace = new LinkedHashSet();
-    preserveSpace = new LinkedHashSet();
-    outputCdataSectionElements = new LinkedHashSet();
+    extensionElementPrefixes = new HashSet<String>();
+    excludeResultPrefixes = new HashSet<String>();
+    stripSpace = new LinkedHashSet<StrippingInstruction>();
+    preserveSpace = new LinkedHashSet<StrippingInstruction>();
+    outputCdataSectionElements = new LinkedHashSet<String>();
     xpath = (XPathImpl) factory.xpathFactory.newXPath();
+    xpath.setNamespaceContext(this);
     if (parent == null)
       {
         bindings = new Bindings(this);
-        attributeSets = new LinkedList();
-        variables = new LinkedList();
-        namespaceAliases = new LinkedHashMap();
-        templates = new LinkedList();
-        keys = new LinkedList();
-        decimalFormats = new LinkedHashMap();
+        attributeSets = new LinkedList<AttributeSet>();
+        variables = new LinkedList<ParameterNode>();
+        namespaceAliases = new LinkedHashMap<String,String>();
+        templates = new LinkedList<Template>();
+        keys = new LinkedList<Key>();
+        decimalFormats = new LinkedHashMap<String,DecimalFormat>();
         initDefaultDecimalFormat();
-        xpath.setNamespaceContext(this);
         xpath.setXPathFunctionResolver(this);
       }
     else
@@ -241,13 +244,12 @@ class Stylesheet
         templates = root.templates;
         keys = root.keys;
         decimalFormats = root.decimalFormats;
-        xpath.setNamespaceContext(root);
         xpath.setXPathFunctionResolver(root);
       }
     xpath.setXPathVariableResolver(bindings);
 
     Test anyNode = new NodeTypeTest((short) 0);
-    List tests = Collections.singletonList(anyNode);
+    List<Test> tests = Collections.singletonList(anyNode);
     builtInNodeTemplate =
       new ApplyTemplatesNode(new Selector(Selector.CHILD, tests),
                              null, null, null, true);
@@ -263,9 +265,8 @@ class Stylesheet
     if (debug)
       {
         System.err.println("Stylesheet: " + doc.getDocumentURI());
-        for (Iterator i = templates.iterator(); i.hasNext(); )
+	for (Template t : templates)
           {
-            Template t = (Template) i.next();
             t.list(System.err);
             System.err.println("--------------------");
           }
@@ -276,9 +277,7 @@ class Stylesheet
   {
     Stylesheet stylesheet = this;
     while (stylesheet.parent != null)
-      {
-        stylesheet = stylesheet.parent;
-      }
+      stylesheet = stylesheet.parent;
     return stylesheet;
   }
   
@@ -309,34 +308,30 @@ class Stylesheet
         Stylesheet clone = (Stylesheet) super.clone();
         clone.bindings = (Bindings) bindings.clone();
 
-        LinkedList templates2 = new LinkedList();
-        for (Iterator i = templates.iterator(); i.hasNext(); )
+        LinkedList<Template> templates2 = new LinkedList<Template>();
+	for (Template t : templates)
           {
-            Template t = (Template) i.next();
             templates2.add(t.clone(clone));
           }
         clone.templates = templates2;
 
-        LinkedList attributeSets2 = new LinkedList();
-        for (Iterator i = attributeSets.iterator(); i.hasNext(); )
+        LinkedList<AttributeSet> attributeSets2 = new LinkedList<AttributeSet>();
+	for (AttributeSet as : attributeSets)
           {
-            AttributeSet as = (AttributeSet) i.next();
             attributeSets2.add(as.clone(clone));
           }
         clone.attributeSets = attributeSets2;
 
-        LinkedList variables2 = new LinkedList();
-        for (Iterator i = variables.iterator(); i.hasNext(); )
+        LinkedList<ParameterNode> variables2 = new LinkedList<ParameterNode>();
+	for (ParameterNode var : variables)
           {
-            ParameterNode var = (ParameterNode) i.next();
             variables2.add(var.clone(clone));
           }
         clone.variables = variables2;
 
-        LinkedList keys2 = new LinkedList();
-        for (Iterator i = keys.iterator(); i.hasNext(); )
+        LinkedList<Key> keys2 = new LinkedList<Key>();
+	for (Key k : keys)
           {
-            Key k = (Key) i.next();
             keys2.add(k.clone(clone));
           }
         clone.keys = keys2;
@@ -359,11 +354,10 @@ class Stylesheet
     // See XSLT 11.4: "If the template or expression specifying the value of
     // a global variable x references a global variable y, then the value
     // for y must be computed before the value of x."
-    List topLevel = new ArrayList(variables);
+    List<ParameterNode> topLevel = new ArrayList<ParameterNode>(variables);
     Collections.sort(topLevel);
-    for (Iterator i = topLevel.iterator(); i.hasNext(); )
+    for (ParameterNode var : topLevel)
       {
-        ParameterNode var = (ParameterNode) i.next();
         bindings.set(var.name,
                      var.getValue(this, null, context, 1, 1),
                      var.type);
@@ -383,7 +377,7 @@ class Stylesheet
     return (current == null) ? null : current.lookupPrefix(namespaceURI);
   }
 
-  public Iterator getPrefixes(String namespaceURI)
+  public Iterator<String> getPrefixes(String namespaceURI)
   {
     // TODO
     return Collections.singleton(getPrefix(namespaceURI)).iterator();
@@ -408,13 +402,10 @@ class Stylesheet
     throws TransformerException
   {
     if (debug)
+      System.err.println("getTemplate: mode="+mode+" context="+context);
+    Template selected = null;
+    for (Template t : templates)
       {
-        System.err.println("getTemplate: mode="+mode+" context="+context);
-      }
-    Set candidates = new TreeSet();
-    for (Iterator j = templates.iterator(); j.hasNext(); )
-      {
-        Template t = (Template) j.next();
         boolean isMatch = t.matches(mode, context);
         if (applyImports)
           {
@@ -425,25 +416,30 @@ class Stylesheet
                 throw new TransformerException(msg);
               }
             if (!currentTemplate.imports(t))
-              {
-                isMatch = false;
-              }
+              isMatch = false;
           }
         //System.err.println("\t"+context+" "+t+"="+isMatch);
         if (isMatch)
           {
-            candidates.add(t);
+            // Conflict resolution
+            // @see http://www.w3.org/TR/xslt#conflict
+            if (selected == null)
+              selected = t;
+            else
+              {
+                if (t.precedence < selected.precedence ||
+                    t.priority < selected.priority)
+                  continue;
+                selected = t;
+              }
           }
       }
-    //System.err.println("\tcandidates="+candidates);
-    if (candidates.isEmpty())
+    if (selected == null)
       {
         // Apply built-in template
         // Current template is unchanged
         if (debug)
-          {
-            System.err.println("\tbuiltInTemplate context="+context);
-          }
+          System.err.println("\tbuiltInTemplate context="+context);
         switch (context.getNodeType())
           {
           case Node.ELEMENT_NODE:
@@ -453,48 +449,45 @@ class Stylesheet
           case Node.COMMENT_NODE:
             return builtInNodeTemplate;
           case Node.TEXT_NODE:
+          case Node.CDATA_SECTION_NODE:
           case Node.ATTRIBUTE_NODE:
             return builtInTextTemplate;
           default:
             return null;
           }
       }
-    else
-      {
-        Template t = (Template) candidates.iterator().next();
-        // Set current template
-        currentTemplate = t;
-        if (debug)
-          {
-            System.err.println("\ttemplate="+t+" context="+context);
-          }
-        return t.node;
-      }
+    // Set current template
+    currentTemplate = selected;
+    if (debug)
+      System.err.println("\ttemplate="+currentTemplate+" context="+context);
+    return currentTemplate.node;
   }
 
   TemplateNode getTemplate(QName mode, QName name)
     throws TransformerException
   {
-    //System.err.println("getTemplate: mode="+mode+" name="+name);
-    Set candidates = new TreeSet();
-    for (Iterator j = templates.iterator(); j.hasNext(); )
+    Template selected = null;
+    for (Template t : templates)
       {
-        Template t = (Template) j.next();
         boolean isMatch = t.matches(name);
-        //System.err.println("\t"+name+" "+t+"="+isMatch);
         if (isMatch)
           {
-            candidates.add(t);
+            // Conflict resolution
+            // @see http://www.w3.org/TR/xslt#conflict
+            if (selected == null)
+              selected = t;
+            else
+              {
+                if (t.precedence < selected.precedence ||
+                    t.priority < selected.priority)
+                  continue;
+                selected = t;
+              }
           }
       }
-    if (candidates.isEmpty())
-      {
-        return null;
-        //throw new TransformerException("template '" + name + "' not found");
-      }
-    Template t = (Template) candidates.iterator().next();
-    //System.err.println("\ttemplate="+t+" context="+context);
-    return t.node;
+    if (selected == null)
+      return null;
+    return selected.node;
   }
 
   /**
@@ -522,11 +515,9 @@ class Stylesheet
     String p = getAttribute(attrs, "priority");
     String mm = getAttribute(attrs, "mode");
     QName mode = (mm == null) ? null : getQName(mm);
-    double priority = (p == null) ? Template.DEFAULT_PRIORITY :
-      Double.parseDouble(p);
     Node children = node.getFirstChild();
     return new Template(this, name, match, parse(children),
-                        precedence, priority, mode);
+                        precedence, p, mode);
   }
 
   /**
@@ -538,42 +529,30 @@ class Stylesheet
     output = node;
     String method = getAttribute(attrs, "method");
     if ("xml".equals(method) || method == null)
-      {
-        outputMethod = OUTPUT_XML;
-      }
+      outputMethod = OUTPUT_XML;
     else if ("html".equals(method))
-      {
-        outputMethod = OUTPUT_HTML;
-      }
+      outputMethod = OUTPUT_HTML;
     else if ("text".equals(method))
-      {
-        outputMethod = OUTPUT_TEXT;
-      }
+      outputMethod = OUTPUT_TEXT;
     else
       {
         String msg = "unsupported output method: " + method;
         DOMSourceLocator l = new DOMSourceLocator(node);
         throw new TransformerConfigurationException(msg, l);
       }
-    outputPublicId = getAttribute(attrs, "public-id");
-    outputSystemId = getAttribute(attrs, "system-id");
+    outputPublicId = getAttribute(attrs, "doctype-public");
+    outputSystemId = getAttribute(attrs, "doctype-system");
     outputEncoding = getAttribute(attrs, "encoding");
     String indent = getAttribute(attrs, "indent");
     if (indent != null)
-      {
-        outputIndent = "yes".equals(indent);
-      }
+      outputIndent = "yes".equals(indent);
     outputVersion = getAttribute(attrs, "version");
     String omitXmlDecl = getAttribute(attrs, "omit-xml-declaration");
     if (omitXmlDecl != null)
-      {
-        outputOmitXmlDeclaration = "yes".equals(omitXmlDecl);
-      }
+      outputOmitXmlDeclaration = "yes".equals(omitXmlDecl);
     String standalone = getAttribute(attrs, "standalone");
     if (standalone != null)
-      {
-        outputStandalone = "yes".equals(standalone);
-      }
+      outputStandalone = "yes".equals(standalone);
     outputMediaType = getAttribute(attrs, "media-type");
     String cdataSectionElements =
       getAttribute(attrs, "cdata-section-elements");
@@ -581,9 +560,7 @@ class Stylesheet
       {
         StringTokenizer st = new StringTokenizer(cdataSectionElements, " ");
         while (st.hasMoreTokens())
-          {
-            outputCdataSectionElements.add(st.nextToken());
-          }
+          outputCdataSectionElements.add(st.nextToken());
       }
   }
 
@@ -728,9 +705,7 @@ class Stylesheet
                 parse(node.getFirstChild(), false);
               }
             else if ("template".equals(name))
-              {
-                templates.add(parseTemplate(node, attrs));
-              }
+              templates.add(parseTemplate(node, attrs));
             else if ("param".equals(name) ||
                      "variable".equals(name))
               {
@@ -779,9 +754,7 @@ class Stylesheet
                 factory.newStylesheet(source, precedence + delta, this);
               }
             else if ("output".equals(name))
-              {
-                parseOutput(node, attrs);
-              }
+              parseOutput(node, attrs);
             else if ("preserve-space".equals(name))
               {
                 String elements =
@@ -790,7 +763,9 @@ class Stylesheet
                                                          " \t\n\r");
                 while (st.hasMoreTokens())
                   {
-                    preserveSpace.add(parseNameTest(st.nextToken()));
+                    NameTest element = parseNameTest(st.nextToken());
+                    preserveSpace.add(new StrippingInstruction(element,
+                                                               precedence));
                   }
               }
             else if ("strip-space".equals(name))
@@ -801,25 +776,19 @@ class Stylesheet
                                                          " \t\n\r");
                 while (st.hasMoreTokens())
                   {
-                    stripSpace.add(parseNameTest(st.nextToken()));
+                    NameTest element = parseNameTest(st.nextToken());
+                    stripSpace.add(new StrippingInstruction(element, 
+                                                            precedence));
                   }
               }
             else if ("key".equals(name))
-              {
-                parseKey(node, attrs);
-              }
+              parseKey(node, attrs);
             else if ("decimal-format".equals(name))
-              {
-                parseDecimalFormat(node, attrs);
-              }
+              parseDecimalFormat(node, attrs);
             else if ("namespace-alias".equals(name))
-              {
-                parseNamespaceAlias(node, attrs);
-              }
+              parseNamespaceAlias(node, attrs);
             else if ("attribute-set".equals(name))
-              {
-                parseAttributeSet(node, attrs);
-              }
+              parseAttributeSet(node, attrs);
           }
         else if (root)
           {
@@ -839,7 +808,7 @@ class Stylesheet
             templates.add(new Template(this, null, new Root(),
                                        parse(rootClone),
                                        precedence,
-                                       Template.DEFAULT_PRIORITY,
+                                       null,
                                        null));
           }
         else
@@ -867,12 +836,10 @@ class Stylesheet
   final NameTest parseNameTest(String token)
   {
     if ("*".equals(token))
-      {
-        return new NameTest(null, true, true);
-      }
+      return new NameTest(null, true, true);
     else if (token.endsWith(":*"))
       {
-        QName qName = getQName(token.substring(0, token.length() - 2));
+        QName qName = getQName(token);
         return new NameTest(qName, true, false);
       }
     else
@@ -889,8 +856,8 @@ class Stylesheet
     // Tokenize
     int len = value.length();
     int off = 0;
-    List tokens = new ArrayList(); // text tokens
-    List types = new ArrayList(); // literal or expression
+    List<String> tokens = new ArrayList<String>(); // text tokens
+    List<Boolean> types = new ArrayList<Boolean>(); // literal or expression
     int depth = 0;
     for (int i = 0; i < len; i++)
       {
@@ -963,8 +930,8 @@ class Stylesheet
     len = tokens.size();
     for (int i = len - 1; i >= 0; i--)
       {
-        String token = (String) tokens.get(i);
-        Boolean type = (Boolean) types.get(i);
+        String token = tokens.get(i);
+        Boolean type = types.get(i);
         if (type == Boolean.TRUE)
           {
             // Expression text
@@ -984,7 +951,13 @@ class Stylesheet
     return ret;
   }
 
-  boolean isPreserved(Text text)
+  /**
+   * Whitespace stripping.
+   * @param text the text node
+   * @param source true if a source node, false if a stylesheet text node
+   * @see http://www.w3.org/TR/xslt#strip
+   */
+  boolean isPreserved(Text text, boolean source)
     throws TransformerConfigurationException
   {
     // Check characters in text
@@ -996,39 +969,71 @@ class Stylesheet
           {
             char c = value.charAt(i);
             if (c != 0x20 && c != 0x09 && c != 0x0a && c != 0x0d)
-              {
-                return true;
-              }
+              return true;
           }
       }
     // Check parent node
     Node ctx = text.getParentNode();
-    if (!preserveSpace.isEmpty())
+    if (source)
       {
-        for (Iterator i = preserveSpace.iterator(); i.hasNext(); )
+        // Source document text node
+        boolean preserve = true;
+        float psPriority = 0.0f, ssPriority = 0.0f;
+        if (!stripSpace.isEmpty())
           {
-            NameTest preserveTest = (NameTest) i.next();
-            if (preserveTest.matches(ctx, 1, 1))
+            // Conflict resolution
+            StrippingInstruction ssi = null, psi = null;
+	    for (StrippingInstruction si : stripSpace)
               {
-                boolean override = false;
-                if (!stripSpace.isEmpty())
+                if (si.element.matches(ctx, 1, 1))
                   {
-                    for (Iterator j = stripSpace.iterator(); j.hasNext(); )
+                    if (ssi != null)
                       {
-                        NameTest stripTest = (NameTest) j.next();
-                        if (stripTest.matches(ctx, 1, 1))
-                          {
-                            override = true;
-                            break;
-                          }
+                        if (si.precedence < ssi.precedence)
+                          continue;
+                        float p = si.getPriority();
+                        if (p < ssPriority)
+                          continue;
                       }
-                  }
-                if (!override)
-                  {
-                    return true;
+                    ssi = si;
                   }
               }
+	    for (StrippingInstruction si : preserveSpace)
+              {
+                if (si.element.matches(ctx, 1, 1))
+                  {
+                    if (psi != null)
+                      {
+                        if (si.precedence < psi.precedence)
+                          continue;
+                        float p = si.getPriority();
+                        if (p < psPriority)
+                          continue;
+                      }
+                    psi = si;
+                  }
+              }
+            if (ssi != null)
+              {
+                if (psi != null)
+                  {
+                    if (psi.precedence < ssi.precedence)
+                      preserve = false;
+                    else if (psPriority < ssPriority)
+                      preserve = false;
+                  }
+                else
+                  preserve = false;
+              }
           }
+        if (preserve)
+          return true;
+      }
+    else
+      {
+        // Stylesheet text node
+        if (STYLESHEET_PRESERVE_TEXT.matches(ctx, 1, 1))
+          return true;
       }
     // Check whether any ancestor specified xml:space
     while (ctx != null)
@@ -1038,24 +1043,14 @@ class Stylesheet
             Element element = (Element) ctx;
             String xmlSpace = element.getAttribute("xml:space");
             if ("default".equals(xmlSpace))
-              {
-                break;
-              }
+              break;
             else if ("preserve".equals(xmlSpace))
-              {
-                return true;
-              }
+              return true;
             else if (xmlSpace.length() > 0)
               {
                 String msg = "Illegal value for xml:space: " + xmlSpace;
                 throw new TransformerConfigurationException(msg);
               }
-            else if ("text".equals(ctx.getLocalName()) &&
-                     XSL_NS.equals(ctx.getNamespaceURI()))
-              {
-                // xsl:text implies xml:space='preserve'
-                return true;
-              }
           }
         ctx = ctx.getParentNode();
       }
@@ -1071,45 +1066,27 @@ class Stylesheet
         if ("document".equals(localName) && (arity == 1 || arity == 2))
           {
             if (current == null)
-              {
                 throw new RuntimeException("current is null");
-              }
             return new DocumentFunction(getRootStylesheet(), current);
           }
         else if ("key".equals(localName) && (arity == 2))
-          {
-            return new KeyFunction(getRootStylesheet());
-          }
+          return new KeyFunction(getRootStylesheet());
         else if ("format-number".equals(localName) &&
                  (arity == 2 || arity == 3))
-          {
-            return new FormatNumberFunction(getRootStylesheet());
-          }
+          return new FormatNumberFunction(getRootStylesheet());
         else if ("current".equals(localName) && (arity == 0))
-          {
-            return new CurrentFunction(getRootStylesheet());
-          }
+          return new CurrentFunction(getRootStylesheet());
         else if ("unparsed-entity-uri".equals(localName) && (arity == 1))
-          {
-            return new UnparsedEntityUriFunction();
-          }
+          return new UnparsedEntityUriFunction();
         else if ("generate-id".equals(localName) &&
                  (arity == 1 || arity == 0))
-          {
-            return new GenerateIdFunction();
-          }
+          return new GenerateIdFunction();
         else if ("system-property".equals(localName) && (arity == 1))
-          {
-            return new SystemPropertyFunction();
-          }
+          return new SystemPropertyFunction();
         else if ("element-available".equals(localName) && (arity == 1))
-          {
-            return new ElementAvailableFunction(this);
-          }
+          return new ElementAvailableFunction(new NamespaceProxy(current));
         else if ("function-available".equals(localName) && (arity == 1))
-          {
-            return new FunctionAvailableFunction(this);
-          }
+          return new FunctionAvailableFunction(new NamespaceProxy(current));
       }
     return null;
   }
@@ -1127,12 +1104,10 @@ class Stylesheet
     QName mode = (m == null) ? null : getQName(m);
     String s = getAttribute(attrs, "select");
     if (s == null)
-      {
-        s = "child::node()";
-      }
+      s = "child::node()";
     Node children = node.getFirstChild();
-    List sortKeys = parseSortKeys(children);
-    List withParams = parseWithParams(children);
+    List<SortKey> sortKeys = parseSortKeys(children);
+    List<WithParam> withParams = parseWithParams(children);
     Expr select = (Expr) xpath.compile(s);
     return new ApplyTemplatesNode(select, mode,
                                   sortKeys, withParams, false);
@@ -1148,7 +1123,7 @@ class Stylesheet
     String n = getRequiredAttribute(attrs, "name", node);
     QName name = getQName(n);
     Node children = node.getFirstChild();
-    List withParams = parseWithParams(children);
+    List<WithParam> withParams = parseWithParams(children);
     return new CallTemplateNode(name, withParams);
   }
   
@@ -1175,7 +1150,7 @@ class Stylesheet
     NamedNodeMap attrs = node.getAttributes();
     String s = getRequiredAttribute(attrs, "select", node);
     Node children = node.getFirstChild();
-    List sortKeys = parseSortKeys(children);
+    List<SortKey> sortKeys = parseSortKeys(children);
     Expr select = (Expr) xpath.compile(s);
     ForEachNode ret = new ForEachNode(select, sortKeys);
     ret.children = parse(children);
@@ -1411,13 +1386,9 @@ class Stylesheet
         if (tnode != null)
           {
             if (first == null)
-              {
-                first = tnode;
-              }
+              first = tnode;
             if (previous != null)
-              {
-                previous.next = tnode;
-              }
+              previous.next = tnode;
             previous = tnode;
           }
         node = next;
@@ -1438,25 +1409,15 @@ class Stylesheet
           {
             String name = node.getLocalName();
             if ("apply-templates".equals(name))
-              {
-                return parseApplyTemplates(node);
-              }
+              return parseApplyTemplates(node);
             else if ("call-template".equals(name))
-              {
-                return parseCallTemplate(node);
-              }
+              return parseCallTemplate(node);
             else if ("value-of".equals(name))
-              {
-                return parseValueOf(node);
-              }
+              return parseValueOf(node);
             else if ("for-each".equals(name))
-              {
-                return parseForEach(node);
-              }
+              return parseForEach(node);
             else if ("if".equals(name))
-              {
-                return parseIf(node);
-              }
+              return parseIf(node);
             else if ("choose".equals(name))
               {
                 Node children = node.getFirstChild();
@@ -1465,9 +1426,7 @@ class Stylesheet
                 return ret;
               }
             else if ("when".equals(name))
-              {
-                return parseWhen(node);
-              }
+              return parseWhen(node);
             else if ("otherwise".equals(name))
               {
                 Node children = node.getFirstChild();
@@ -1476,25 +1435,15 @@ class Stylesheet
                 return ret;
               }
             else if ("element".equals(name))
-              {
-                return parseElement(node);
-              }
+              return parseElement(node);
             else if ("attribute".equals(name))
-              {
-                return parseAttribute(node);
-              }
+              return parseAttribute(node);
             else if ("text".equals(name))
-              {
-                return parseText(node);
-              }
+              return parseText(node);
             else if ("copy".equals(name))
-              {
-                return parseCopy(node);
-              }
+              return parseCopy(node);
             else if ("processing-instruction".equals(name))
-              {
-                return parseProcessingInstruction(node);
-              }
+              return parseProcessingInstruction(node);
             else if ("comment".equals(name))
               {
                 Node children = node.getFirstChild();
@@ -1503,9 +1452,7 @@ class Stylesheet
                 return ret;
               }
             else if ("number".equals(name))
-              {
-                return parseNumber(node);
-              }
+              return parseNumber(node);
             else if ("param".equals(name) ||
                      "variable".equals(name))
               {
@@ -1538,13 +1485,9 @@ class Stylesheet
                 return ret;
               }
             else if ("copy-of".equals(name))
-              {
-                return parseCopyOf(node);
-              }
+              return parseCopyOf(node);
             else if ("message".equals(name))
-              {
-                return parseMessage(node);
-              }
+              return parseMessage(node);
             else if ("apply-imports".equals(name))
               {
                 Node children = node.getFirstChild();
@@ -1562,22 +1505,30 @@ class Stylesheet
         String prefix = node.getPrefix();
         if (extensionElementPrefixes.contains(prefix))
           {
-            // Pass over extension elements
+            // Check for xsl:fallback
+            for (Node ctx = node.getFirstChild(); ctx != null;
+                 ctx = ctx.getNextSibling())
+              {
+                String ctxUri = ctx.getNamespaceURI();
+                if (XSL_NS.equals(ctxUri) &&
+                    "fallback".equals(ctx.getLocalName()))
+                  {
+                    ctx = ctx.getFirstChild();
+                    return (ctx == null) ? null : parse(ctx);
+                  }
+              }
+            // Otherwise pass over extension element
             return null;
           }
         switch (node.getNodeType())
           {
           case Node.TEXT_NODE:
+          case Node.CDATA_SECTION_NODE:
             // Determine whether to strip whitespace
             Text text = (Text) node;
-            if (!isPreserved(text))
+            if (!isPreserved(text, false))
               {
                 // Strip
-                /*String data = text.getData().trim();
-                if (data.length() > 0)
-                  {
-                    text.setData(data);
-                  } // else */
                 text.getParentNode().removeChild(text);
                 return null;
               }
@@ -1623,9 +1574,7 @@ class Stylesheet
                     String aname = attr.getNodeName();
                     if (Stylesheet.XSL_NS.equals(ans) &&
                         "use-attribute-sets".equals(attr.getLocalName()))
-                      {
-                        continue;
-                      }
+                      continue;
                     String value = attr.getNodeValue();
                     TemplateNode grandchild =
                       parseAttributeValueTemplate(value, node);
@@ -1640,8 +1589,9 @@ class Stylesheet
                   }
                 String ename = node.getNodeName();
                 TemplateNode n = parseAttributeValueTemplate(ename, node);
-                TemplateNode ns = (namespaceUri == null) ? null :
-                  parseAttributeValueTemplate(namespaceUri, node);
+                //TemplateNode ns = (namespaceUri == null) ? null :
+                //  parseAttributeValueTemplate(namespaceUri, node);
+                TemplateNode ns = null;
                 ElementNode ret = new ElementNode(n, ns, useAttributeSets,
                                                   node);
                 ret.children = child;
@@ -1662,10 +1612,10 @@ class Stylesheet
     return ret;
   }
 
-  final List parseSortKeys(Node node)
+  final List<SortKey> parseSortKeys(Node node)
     throws TransformerConfigurationException, XPathExpressionException
   {
-    List ret = new LinkedList();
+    List<SortKey> ret = new LinkedList<SortKey>();
     while (node != null)
       {
         String namespaceUri = node.getNamespaceURI();
@@ -1676,9 +1626,7 @@ class Stylesheet
             NamedNodeMap attrs = node.getAttributes();
             String s = getAttribute(attrs, "select");
             if (s == null)
-              {
-                s = ".";
-              }
+              s = ".";
             Expr select = (Expr) xpath.compile(s);
             String l = getAttribute(attrs, "lang");
             TemplateNode lang = (l == null) ? null :
@@ -1696,13 +1644,13 @@ class Stylesheet
           }
         node = node.getNextSibling();
       }
-    return ret.isEmpty() ? null : ret;
+    return ret;
   }
 
-  final List parseWithParams(Node node)
+  final List<WithParam> parseWithParams(Node node)
     throws TransformerConfigurationException, XPathExpressionException
   {
-    List ret = new LinkedList();
+    List<WithParam> ret = new LinkedList<WithParam>();
     while (node != null)
       {
         String namespaceUri = node.getNamespaceURI();
@@ -1728,13 +1676,11 @@ class Stylesheet
                 ret.add(new WithParam(name, expr));
               }
             else
-              {
-                ret.add(new WithParam(name, content));
-              }
+              ret.add(new WithParam(name, content));
           }
         node = node.getNextSibling();
       }
-    return ret.isEmpty() ? null : ret;
+    return ret;
   }
 
   /**
@@ -1743,7 +1689,7 @@ class Stylesheet
    * exclude-result-prefixes.
    */
   final void addNamespaceNodes(Node source, Node target, Document doc,
-                               Collection elementExcludeResultPrefixes)
+                               Collection<String> elementExcludeResultPrefixes)
   {
     NamedNodeMap attrs = source.getAttributes();
     if (attrs != null)
@@ -1757,27 +1703,19 @@ class Stylesheet
               {
                 String prefix = attr.getLocalName();
                 if (XMLConstants.XMLNS_ATTRIBUTE.equals(prefix))
-                  {
-                    prefix = "#default";
-                  }
+                  prefix = "#default";
                 String ns = attr.getNodeValue();
                 // Should the namespace be excluded?
                 if (XSL_NS.equals(ns) ||
                     extensionElementPrefixes.contains(prefix) ||
                     elementExcludeResultPrefixes.contains(prefix) ||
                     excludeResultPrefixes.contains(prefix))
-                  {
-                    continue;
-                  }
+                  continue;
                 // Is the namespace already defined on the target?
                 if (prefix == "#default")
-                  {
-                    prefix = null;
-                  }
+                  prefix = null;
                 if (target.lookupNamespaceURI(prefix) != null)
-                  {
-                    continue;
-                  }
+                  continue;
                 attr = attr.cloneNode(true);
                 attr = doc.adoptNode(attr);
                 target.getAttributes().setNamedItemNS(attr);
@@ -1786,23 +1724,17 @@ class Stylesheet
       }
     Node parent = source.getParentNode();
     if (parent != null)
-      {
-        addNamespaceNodes(parent, target, doc, elementExcludeResultPrefixes);
-      }
+      addNamespaceNodes(parent, target, doc, elementExcludeResultPrefixes);
   }
 
   static final String getAttribute(NamedNodeMap attrs, String name)
   {
     Node attr = attrs.getNamedItem(name);
     if (attr == null)
-      {
-        return null;
-      }
+      return null;
     String ret = attr.getNodeValue();
     if (ret.length() == 0)
-      {
-        return null;
-      }
+      return null;
     return ret;
   }
 
@@ -1828,5 +1760,14 @@ class Stylesheet
     dst.setUserData(key, data, this);
   }
 
+  public String toString()
+  {
+    CPStringBuilder b = new CPStringBuilder(getClass().getName());
+    b.append("[templates=");
+    b.append(templates);
+    b.append("]");
+    return b.toString();
+  }
+
 }
 
